; Copyright (c) 2005 Inujima, Masaru <qfwfq@kt.rim.or.jp>
;
; Permission to use, copy, modify, and distribute this software for any
; purpose with or without fee is hereby granted, provided that the above
; copyright notice and this permission notice appear in all copies.
;
; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

;;;; $Id: repl.scm,v 1.12 2005/09/14 05:40:02 qfwfq Exp $
; $Log: repl.scm,v $
; Revision 1.12  2005/09/14 05:40:02  qfwfq
; add copyright notice
;
; Revision 1.11  2005/08/04 07:00:15  qfwfq
; trace
;
; Revision 1.10  2005/07/28 09:44:56  qfwfq
; external module loading
;
; Revision 1.9  2005/07/15 08:48:56  qfwfq
; rowstream
;
; Revision 1.8  2005/07/11 11:08:35  qfwfq
; cursor
;
; Revision 1.7  2005/06/28 11:05:09  qfwfq
; prepare & execute
;
; Revision 1.6  2005/06/13 06:47:50  qfwfq
; trigger support
;
; Revision 1.5  2005/05/26 04:52:55  qfwfq
; rd:spi-execute
;
; Revision 1.4  2005/05/13 08:01:26  qfwfq
; procedure to invoke elog
;
; Revision 1.3  2005/04/15 07:15:04  qfwfq
; support SRF
;
; Revision 1.2  2005/04/11 09:58:24  qfwfq
; common numeric types as args and return
;
; Revision 1.1  2005/04/01 10:32:24  qfwfq
; Initial revision
;

(define rd:reset rp:error)
(define rd:null-value (rp:void))
(define (rd:isnull? x) (eq? x rd:null-value))

(define (rd:yield-error msg)
  (rd:do-or-schedule-elog 20 msg) (rd:reset))
(define (rd:srf-yield-error r msg)
  (rd:suspend-interrupt #t) (rd:free-state-rec r) (rd:yield-error msg))

(define rd:alloc-state-rec #f)
(define rd:free-state-rec #f)
(define rd:record-state #f)
(define rd:recall-state #f)
(let ((recvec (make-vector 0))
      (freelist '()))
  (set! rd:alloc-state-rec
    (lambda ()
      (if (null? freelist)
	(let* ((oldsize (vector-length recvec))
	       (newsize (+ oldsize 10))
	       (newrec (make-vector newsize)))
	  (do ((i 0 (+ i 1))) ((= i oldsize) (set! recvec newrec)) (vector-set! newrec i (vector-ref recvec i)))
	  (do ((i oldsize (+ i 1)) (newfree '() (cons i newfree))) ((= i newsize) (set! freelist newfree)))))
      (let ((r (car freelist))) (set! freelist (cdr freelist)) r)))
  (set! rd:free-state-rec (lambda (r) (vector-set! recvec r #f) (set! freelist (cons r freelist))))
  (set! rd:record-state (lambda (r s) (vector-set! recvec r s)))
  (set! rd:recall-state (lambda (r) (vector-ref recvec r))))

(define (rd:do-spicall task)
  (rd:set-spicall-task task)
  (let ((r (rd:alloc-state-rec)))
    (rd:set-spicall-context r)
    (let ((aborted (call-with-current-continuation
		     (lambda (c)
		       (rd:record-state r c)
		       (rd:reset)))))
      (rd:free-state-rec r)
      (if aborted (rp:error 200)))))

(define rd:emit-message
  (let ((levels '((warning . 19) (notice . 18) (info . 17) (commerror . 16) (log . 15)
		  (debug1 . 14) (debug2 . 13) (debug3 . 12) (debug4 . 11) (debug5 . 10))))
    (lambda (level msg)
      (rd:suspend-interrupt #t)
      (rd:do-or-schedule-elog (cdr (assq level levels)) msg)
      (rd:do-spicall 2)
      (rd:suspend-interrupt #f))))

(define rd:spi-error-connect -1)
(define rd:spi-error-copy -2)
(define rd:spi-error-opunknown -3)
(define rd:spi-error-unconnected -4)
(define rd:spi-error-cursor -5)
(define rd:spi-error-argument -6)
(define rd:spi-error-param -7)
(define rd:spi-error-transaction -8)
(define rd:spi-error-noattribute -9)
(define rd:spi-error-nooutfunc -10)
(define rd:spi-error-typunknown -11)
(define rd:spi-ok-connect 1)
(define rd:spi-ok-finish 2)
(define rd:spi-ok-fetch 3)
(define rd:spi-ok-utility 4)
(define rd:spi-ok-select 5)
(define rd:spi-ok-selinto 6)
(define rd:spi-ok-insert 7)
(define rd:spi-ok-delete 8)
(define rd:spi-ok-update 9)
(define rd:spi-ok-cursor 10)
(define rd:spi-result-error? negative?)
(define rd:spi-result-string
  (let ((ercodes (vector "SPI_ERROR_CONNECT" "SPI_ERROR_COPY" "SPI_ERROR_OPUNKNOWN" "SPI_ERROR_UNCONNECTED"
			 "SPI_ERROR_CURSOR" "SPI_ERROR_ARGUMENT" "SPI_ERROR_PARAM" "SPI_ERROR_TRANSACTION"
			 "SPI_ERROR_NOATTRIBUTE" "SPI_ERROR_NOOUTFUNC" "SPI_ERROR_TYPUNKNOWN"))
	(okcodes (vector "SPI_OK_CONNECT" "SPI_OK_FINISH" "SPI_OK_FETCH" "SPI_OK_UTILITY" "SPI_OK_SELECT"
			 "SPI_OK_SELINTO" "SPI_OK_INSERT" "SPI_OK_DELETE" "SPI_OK_UPDATE" "SPI_OK_CURSOR")))
    (lambda (rc)
      (cond ((< 0 rc 11) (vector-ref okcodes (- rc 1)))
	    ((< -12 rc 0) (vector-ref ercodes (- -1 rc)))
	    (else (string-append "Unrecognized SPI code " (number->string rc)))))))

(define (rd:get-spicall-option defaults opts)
  (if (null? opts) defaults
    (cons (car opts) (rd:get-spicall-option (cdr defaults) (cdr opts)))))

(define rd:get-spi-results
  (let ((vprocs (vector rd:get-attribute-value-bool rd:get-attribute-value-char
			rd:get-attribute-value-int2 rd:get-attribute-value-int4
			rd:get-attribute-value-float4 rd:get-attribute-value-float8
			(lambda (rn fn)
			  (rd:set-attribute-row-column-number rn fn)
			  (rd:do-spicall 4)
			  (rd:get-attribute-value-string)))))
    (define (get-tuptable natts rowcnt)
      (let ((getattr
	      (let loop ((i 0))
		(if (= i natts) '()
		  (let ((proc (vector-ref vprocs (rd:get-attribute-type i))))
		    (cons (lambda (rn) (proc rn i)) (loop (+ i 1)))))))
	    (tupv (make-vector rowcnt)))
	(do ((i 0 (+ i 1))) ((= i rowcnt) tupv)
	  (vector-set! tupv i (map (lambda (p) (p i)) getattr)))))
    (define (get-fnams natts)
      (let loop ((i 0))
	(if (= i natts) '()
	  (cons (rd:get-attribute-name i) (loop (+ i 1))))))
    (lambda (getfnam)
      (let ((rc (rd:get-spi-result-code))
	    (rowcnt (rd:get-spi-processed-rows)))
	(if (or (rd:spi-result-error? rc) (= rc rd:spi-ok-utility))
	  (set! rowcnt rd:null-value))
	(if (= rc rd:spi-ok-select)
	  (let ((natts (rd:get-number-of-attributes)))
	    (let ((tupv (get-tuptable natts rowcnt))
		  (fnams (if getfnam (get-fnams natts))))
	      (rd:free-spi-tuple-table)
	      (rd:suspend-interrupt #f)
	      (if getfnam
		(values rc rowcnt tupv fnams)
		(values rc rowcnt tupv))))
	  (begin (rd:suspend-interrupt #f) (values rc rowcnt)))))))

(define (rd:spi-execute sql . opts)
  (rd:suspend-interrupt #t)
  (apply
    (lambda (cnt getfnam)
      (rd:set-sql-command sql)
      (if (not (rd:check-volatility)) (rp:error 201))
      (rd:set-max-rows cnt)
      (rd:do-spicall 3)
      (rd:get-spi-results getfnam))
    (rd:get-spicall-option '(0 #f) opts)))

(define (rd:spi-prepare sql . args)
  (rd:suspend-interrupt #t)
  (rd:set-sql-command sql)
  (rd:set-plan-argument-types args)
  (rd:do-spicall 5)
  (let ((pobj (rd:make-plan-object)))
    (rd:suspend-interrupt #f)
    pobj))

(define (rd:spi-execute-plan plan params . opts)
  (rd:suspend-interrupt #t)
  (apply
    (lambda (cnt getfnam)
      (rd:set-plan-object plan)
      (do ((i 0 (+ i 1)) (ps params (cdr ps))) ((null? ps))
	(rd:set-plan-parameter i (car ps)))
      (if (not (rd:check-volatility)) (rp:error 201))
      (rd:set-max-rows cnt)
      (rd:do-spicall 6)
      (rd:get-spi-results getfnam))
    (rd:get-spicall-option '(0 #f) opts)))

(define (rd:spi-cursor-open name plan params)
  (rd:suspend-interrupt #t)
  (rd:set-cursor-name name)
  (rd:set-plan-object plan)
  (do ((i 0 (+ i 1)) (ps params (cdr ps))) ((null? ps))
    (rd:set-plan-parameter i (car ps)))
  (if (not (rd:check-volatility)) (rp:error 201))
  (rd:do-spicall 7)
  (let ((cobj (rd:make-cursor-object)))
    (rd:suspend-interrupt #f)
    cobj))

(define (rd:spi-cursor-find name)
  (rd:suspend-interrupt #t)
  (rd:set-cursor-name name)
  (rd:do-spicall 8)
  (let ((cobj (rd:make-cursor-object)))
    (rd:suspend-interrupt #f)
    cobj))

(define (rd:spi-cursor-fetch cursor cnt . opts)
  (rd:suspend-interrupt #t)
  (apply
    (lambda (fwd getfnam)
      (rd:set-cursor-object cursor)
      (rd:set-forward-flag fwd)
      (rd:set-max-rows cnt)
      (rd:do-spicall 9)
      (rd:get-spi-results getfnam))
    (rd:get-spicall-option '(#t #f) opts)))

(define (rd:spi-cursor-move cursor cnt . opts)
  (rd:suspend-interrupt #t)
  (apply
    (lambda (fwd)
      (rd:set-cursor-object cursor)
      (rd:set-forward-flag fwd)
      (rd:set-max-rows cnt)
      (rd:do-spicall 10)
      (rd:suspend-interrupt #f))
    (rd:get-spicall-option '(#t) opts)))

(define (rd:spi-cursor-close cursor)
  (rd:suspend-interrupt #t)
  (rd:set-cursor-object cursor)
  (rd:do-spicall 11)
  (rd:suspend-interrupt #f))

(define (rd:cursor->rowstream cursor)
  (letrec ((mkstrm (lambda ()
		     (delay (call-with-values
			      (lambda () (rd:spi-cursor-fetch cursor 1))
			      (lambda (rc rn rows)
				(if (zero? rn)
				  (begin (rd:spi-cursor-close cursor) '())
				  (cons (vector-ref rows 0) (mkstrm)))))))))
    (mkstrm)))

(define (rd:plan->rowstream plan params)
  (rd:cursor->rowstream (rd:spi-cursor-open rd:null-value plan params)))

(define (rd:sql->rowstream sql)
  (let* ((plan (rd:spi-prepare sql))
	 (stream (rd:plan->rowstream plan '())))
    (rd:spi-freeplan plan)
    stream))

(define rd:module-search-path '())

(define (rd:set-module-search-path . path) (set! rd:module-search-path path))

(define (rd:load-module modname)
  (let ((namelen (string-length modname)))
    (let ((absolute? (char=? (string-ref modname 0) #\/))
	  (compiled? (and (>= namelen 3) (string=? (substring modname (- namelen 3) namelen) ".so")))
	  (source? (and (>= namelen 4) (string=? (substring modname (- namelen 4) namelen) ".scm"))))
      (define (load-if-exists loader file)
	(cond ((file-exists? file) (loader file) (rd:emit-message 'notice (string-append "loaded file " file)) #t)
	      (else #f)))
      (define (try-compiled file)
	(cond (compiled? (load-if-exists rp:load-compiled-module file))
	      (source? #f)
	      (else (load-if-exists rp:load-compiled-module (string-append file ".so")))))
      (define (try-source file)
	(cond (compiled? #f)
	      (source? (load-if-exists load file))
	      (else (load-if-exists load (string-append file ".scm")))))
      (define (tryload file) (or (try-compiled file) (try-source file)))
      (define (pathlist str)
	(let ((len (string-length str)))
	  (if (zero? len) rd:module-search-path
	    (let spos ((j 0))
	      (let epos ((k j))
		(cond ((= k len) (list (substring str j k)))
		      ((char=? (string-ref str k) #\:) (cons (substring str j k) (spos (+ k 1))))
		      (else (epos (+ k 1)))))))))
      (if absolute?
	(or (tryload modname) (rd:emit-message 'warning (string-append "file " modname " not found")))
	(let search ((path (cond ((rd:conf-module-search-path) => pathlist)
				 (else rd:module-search-path))))
	  (cond ((null? path) (rd:emit-message 'warning (string-append "module " modname " not found")))
		((or (zero? (string-length (car path))) (not (char=? (string-ref (car path) 0) #\/)))
		 (rd:emit-message 'warning (string-append "ignoring non absolute path element " (car path)))
		 (search (cdr path)))
		((tryload (string-append (car path) "/" modname)))
		(else (search (cdr path)))))))))

(define rd:trace-tag-list
  (delay
    (let* ((str (rd:conf-trace-tag-list))
	   (len (if str (string-length str) 0)))
      (if (zero? len) '()
	(let spos ((j 0))
	  (let epos ((k j))
	    (cond ((= k len) (list (string->symbol (substring str j k))))
		  ((char=? (string-ref str k) #\,) (cons (string->symbol (substring str j k)) (spos (+ k 1))))
		  (else (epos (+ k 1))))))))))

(define rd:trace-output
  (let ((depth 0))
    (lambda (nam args exp)
      (let ((p (rp:open-output-string)))
	(do ((n depth (- n 1))) ((negative? n) (display #\space p)) (display #\> p))
	(write (cons nam args) p)
	(rd:emit-message 'debug1 (rp:get-output-string p))
	(close-output-port p))
      (call-with-values
	(lambda ()
	  (let ((odepth depth) (ndepth (+ depth 1)))
	    (dynamic-wind (lambda () (set! depth ndepth)) exp (lambda () (set! depth odepth)))))
	(lambda vals
	  (let ((p (rp:open-output-string)))
	    (do ((n depth (- n 1))) ((negative? n)) (display #\< p))
	    (for-each (lambda (v) (display #\space p) (write v p)) vals)
	    (rd:emit-message 'debug1 (rp:get-output-string p))
	    (close-output-port p))
	  (apply values vals))))))

(define-syntax trace-exp
  (lambda (x)
    (syntax-case x ()
      ((_ tag args exp)
       (identifier? (syntax tag))
       (syntax (trace-exp (tag tag) args exp)))
      ((_ (tag label) (arg ...) exp)
       (if (memq (syntax-object->datum (syntax tag)) (force rd:trace-tag-list))
	 (syntax (rd:trace-output 'label (list arg ...) (lambda () exp)))
	 (syntax exp))))))

(define (rd:srf-add-row r v)
  (call-with-current-continuation
    (lambda (c)
      (rd:record-state r c)
      (rd:set-srf-done #f)
      (rd:store-preocedure-return-value v)
      (rd:reset))))

(define (rd:trigger-fired-by-insert? td) (zero? (rp:bitwise-and (vector-ref td 0) 3)))
(define (rd:trigger-fired-by-delete? td) (= (rp:bitwise-and (vector-ref td 0) 3) 1))
(define (rd:trigger-fired-by-update? td) (= (rp:bitwise-and (vector-ref td 0) 3) 2))
(define (rd:trigger-fired-for-row? td) (positive? (rp:bitwise-and (vector-ref td 0) 4)))
(define (rd:trigger-fired-for-statement? td) (zero? (rp:bitwise-and (vector-ref td 0) 4)))
(define (rd:trigger-fired-before? td) (positive? (rp:bitwise-and (vector-ref td 0) 8)))
(define (rd:trigger-fired-after? td) (zero? (rp:bitwise-and (vector-ref td 0) 8)))
(define (rd:trigger-relation-oid td) (vector-ref td 1))
(define (rd:trigger-relation-name td) (vector-ref td 2))
(define (rd:trigger-old-tuple td) (vector-ref td 3))
(define (rd:trigger-new-tuple td) (vector-ref td 4))

(define (rd:process-requests)
  (define src (rd:open-procedure-source))
  (define (intr) (rd:reset))
  (define (do-read-eval)
    (define (check-for-garbage)
      (let ((ch (read-char src)))
	(cond ((eof-object? ch))
	      ((char-whitespace? ch) (check-for-garbage))
	      (else (rd:yield-error "procedure definition has trailing garbage")))))
    (rd:suspend-interrupt #f)
    (let ((def (read src)))
      (check-for-garbage)
      (rd:store-eval-result (eval def))))
  (define (do-apply)
    (rd:suspend-interrupt #f)
    (rd:store-preocedure-return-value
      (apply (rd:get-called-procedure) (rd:get-procedure-arguments))))
  (define (do-srf-first)
    (let ((r (rd:alloc-state-rec)))
      (dynamic-wind
	(lambda ()
	  (rd:set-interrupts-handler (lambda () (rd:free-state-rec r) (rd:set-interrupts-handler intr) (rd:reset)))
	  (rd:suspend-interrupt #f))
	(lambda ()
	  (rd:set-srf-continuation r)
	  (rp:catch-error (lambda (err obj) (rd:suspend-interrupt #t) (rd:free-state-rec r) (rp:error err obj))
	    (apply (rd:get-called-procedure) r (rd:get-procedure-arguments)))
	  (rd:set-srf-done #t))
	(lambda ()
	  (rd:suspend-interrupt #t)
	  (rd:set-interrupts-handler intr)))
      (rd:free-state-rec r)))
  (define (do-srf-next) ((rd:recall-state (rd:get-srf-continuation)) rd:null-value))
  (define (do-spi-return) ((rd:recall-state (rd:get-spicall-context)) #f))
  (define (do-spi-abort) ((rd:recall-state (rd:get-spicall-context)) #t))
  (define (do-trigger)
    (let ((event (rd:get-trigger-event))
	  (relid (rd:get-trigger-relation-oid))
	  (relname (rd:get-trigger-relation-name))
	  (natts (rd:get-trigger-relation-attribute-number))
	  (trigname (rd:get-trigger-trigger-name))
	  (trigargc (rd:get-trigger-trigger-argc)))
      (let ((trigop (rp:bitwise-and event 3))
	    (trigrow? (positive? (rp:bitwise-and event 4)))
	    (trigbefore? (positive? (rp:bitwise-and event 8)))
	    (trigargs (let loop ((k 0))
			(if (= k trigargc) '()
			  (cons (rd:get-trigger-trigger-arg k) (loop (+ k 1))))))
	    (gettuple (lambda (which)
			(let loop ((k 0))
			  (if (= k natts) '()
			    (cons (rd:get-trigger-tuple-attribute which k) (loop (+ k 1))))))))
	(let ((old (if (and trigrow? (positive? trigop)) (gettuple #f)))
	      (new (if trigrow?
		     (cond ((= trigop 0) (gettuple #f))
			   ((= trigop 2) (gettuple #t))))))
	  (rd:suspend-interrupt #f)
	  (let ((retv (apply (rd:get-called-procedure) (vector event relid relname old new) trigname trigargs)))
	    (if (and trigbefore? trigrow?)
	      (cond ((eq? retv #t) (rd:set-trigger-action 0))
		    ((eq? retv #f) (rd:set-trigger-action 1))
		    (else (rd:set-trigger-action 2)
			  (do ((i 0 (+ i 1)) (modatts retv (cdr modatts)))
			    ((= i natts) (if (not (eq? modatts '())) (rp:error 104)))
			    (rd:set-trigger-modified-attribute i (car modatts)))))))))))
  (rd:set-interrupts-handler intr)
  (let ((tasks (vector do-read-eval do-apply do-srf-first do-srf-next do-spi-return do-spi-abort do-trigger)))
    (call-with-current-continuation
      (lambda (c) (set! rd:reset (lambda () (rd:suspend-interrupt #t) (c rd:null-value)))))
    (rp:catch-error
      (lambda (err obj)
	(case err
	  ((200) (rd:set-spicall-task 1) (rd:reset))
	  ((201) (rd:yield-error "database access from immutable function is not permitted"))
	  (else (rd:yield-error (string-append "scheme error: " (rp:error-message err obj))))))
      (begin
	(rd:wait-next-call)
	((vector-ref tasks (rd:get-current-task)))
	(rd:reset)))))

(rd:process-requests)
